package at.ac.uibk.gitsearch.service.dto;

import com.fasterxml.jackson.annotation.JsonInclude;
import java.util.Objects;
import java.util.regex.Pattern;
import javax.validation.constraints.Email;
import javax.validation.constraints.Size;

/**
 * DTO representing users contained in metadata such
 * as creators, publishers, contributors
 */
@JsonInclude(JsonInclude.Include.NON_NULL)
public class MetadataUserDTO {

    @Size(max = 254)
    private String name;

    @Size(max = 254)
    private String affiliation;

    @Email
    @Size(min = 5, max = 254)
    private String email;

    public MetadataUserDTO() {}

    public MetadataUserDTO(String name, String affiliation, String email) {
        this.name = name;
        this.affiliation = affiliation;
        this.email = email;
    }

    public String getName() {
        return name;
    }

    public void setName(String name) {
        this.name = name;
    }

    public String getAffiliation() {
        // Use email domain as fall-back affiliation
        return Objects.requireNonNullElseGet(affiliation, () -> getEmail().substring(getEmail().lastIndexOf("@") + 1));
    }

    public void setAffiliation(String affiliation) {
        this.affiliation = affiliation;
    }

    public String getEmail() {
        return email;
    }

    public void setEmail(String email) {
        this.email = email;
    }

    @Override
    public String toString() {
        return "MetadataUserDTO{" + "name='" + name + '\'' + ", affiliation='" + affiliation + '\'' + ", email='" + email + '\'' + '}';
    }

    public boolean isValid() {
        String emailRegex = "^[A-Za-z0-9+_.-]+@(.+)$";

        Pattern emailPattern = Pattern.compile(emailRegex);

        return (
            !this.getName().isBlank() &&
            !this.getAffiliation().isBlank() &&
            !this.getEmail().isBlank() &&
            emailPattern.matcher(this.getEmail()).matches()
        );
    }

    /**
     * Basic sanitization method replacing brackets with its corresponding escaped symbol
     *
     * @return the sanitized MetadataUserDTO
     */
    public MetadataUserDTO sanitize() {
        this.affiliation = replaceTagCharacters(this.affiliation);
        this.name = replaceTagCharacters(this.name);
        this.email = replaceTagCharacters(this.email);
        return this;
    }

    private String replaceTagCharacters(String input) {
        return input.replace("<", "&lt").replace(">", "&gt");
    }
}
